/*===========================================================================*\

   $Id$

   Copyright (C) 2007 Michael A. Muller

   Permission is granted to use, modify and redistribute this code,
   providing that the following conditions are met:
 
   1) This copyright/licensing notice must remain intact.
   2) If the code is modified and redistributed, the modifications must 
   include documentation indicating that the code has been modified.
   3) The author(s) of this code must be indemnified and held harmless
   against any damage resulting from the use of this code.

   This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
   warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

\*===========================================================================*/

#ifndef COLLIDABLE_H
#define COLLIDABLE_H

#include <spug/RCBase.h>
#include <spug/RCPtr.h>
#include "Vector.h"
#include "Poly.h"
#include "AxisBox.h"

SPUG_RCPTR(Collidable);

/**
 * A "Collidable" is an object that can be checked for a collision by a
 * collision detection algorithm.
 *
 * There are a fixed set of collidable types, any number of which can be
 * expressed by a single collidable.
 */
class Collidable : public virtual spug::RCBase {
   public:
      enum TypeFlag { 

         // if this is in the types, the collidable may be expressed by an
         // AxisBox boundary.  Pretty much everything should satisfy this
         // constraint.
         axisBox = 1,

         // the collidable may be expressed as a sphere.
         sphere = 2,

         // the collidable may be expressed as a polygon
         poly = 4
      };

      /**
       * A set of TypeFlag's combined into a single register.
       */
      class Types {

         private:
            unsigned val;

         private:
            explicit Types(unsigned val) : val(val) {}

         public:
            Types(TypeFlag t) : val(t) {}
            Types operator |(Types other) const { 
               return Types(val | other.val); 
            }
            Types &operator |=(TypeFlag flag) { val |= flag; return *this; }
            bool includes(TypeFlag flag) const { return val & flag; }
      };

      /** Defines a sphere. */
      struct Sphere {
         Vector center;
         Float radius;

         /** Constructs a sphere from its center and radius */
         Sphere(const Vector &center, Float radius) : 
            center(center), 
            radius(radius) {
         }

         /** Default constructor */
         Sphere() : radius(0) {}
      };

   public:
      /**
       * Returns the set of types of this collidable.
       */
      virtual Types getTypes() const = 0;

      /**
       * Returns the priority of the collidable's collision algorithm.  A
       * priority of zero means that collidable has no algorithm, and this
       * checkForCollision() should not be called on it.
       */
      virtual int getPriority() const = 0;

      /**
       * Checks for a collision with the other collidable.  If there is one,
       * returns true and sets "time" to the time till the collision (in
       * seconds).
       */
      virtual bool checkForCollision(const CollidablePtr &other, 
                                     Float &time
                                     ) const = 0;

      /**
       * Used to inform the object that a collision has been detected with the
       * other collidable at a given distance.
       */
      virtual void informOfCollision(const CollidablePtr &other,
				     Float distance
				     ) = 0;

      /**
       * Returns the polygon representation of the object.
       *
       * May be overriden by derived class, base class implementation always
       * throws UnsupportedOperationException.
       */
      virtual const Poly &getPoly() const;

      /**
       * Returns the AxisBox representation of the object.
       *
       * May be overriden by derived class, base class implementation always
       * throws UnsupportedOperationException.
       */
      virtual AxisBox getAxisBox() const;

      /**
       * Returns the Sphere implementation of the object.
       *
       * May be overriden by derived class, base class implementation always
       * throws UnsupportedOperationException.
       */
      virtual const Sphere &getSphere() const;

      /**
       * Returns the object's velocity.
       *
       * May be overriden by derived class, base class implementation returns
       * a zero vector.
       */
      virtual Vector getVelocity() const;

};

std::ostream &operator <<(std::ostream &out, const Collidable &coll);

#endif
