
#include "Moveable.h"
#include "CollisionAlgo.h"
#include <spug/DoesNotExistException.h>

using namespace std;

Moveable::Moveable() : position(0, 0, 0), trace(false) {
   rotation.makeYaw(0);
}

Moveable::Moveable(ModelPtr model, const Vector &position, 
		   const RotMatrix &rotation
	           ) :
   model(model),
   position(position),
   rotation(rotation),
   trace(false) {
}

Moveable::~Moveable() {
   for (CollidableVec::iterator iter = collidables.begin();
        iter != collidables.end();
        ++iter
        )
      (*iter)->setMoveable(0);
}

void Moveable::render(Scene &scene) {

   // if there's no model, quit now.
   if (!model)
      return;

   // store the current position/rotation
   glPushMatrix();

   // translate the rotation matrix 
   Float rotMatrix[4][4];;
   rotation.getGLMatrix(rotMatrix);

   // translate and rotate the prior to rendering
   glLoadIdentity();
   glTranslate(position.x, position.y, position.z);
   glMultMatrix<Float>(rotMatrix);

   // render the model
   model->render(scene);

   // restore the original position
   glPopMatrix();
}

void Moveable::addAlgo(const MoveableAlgoPtr &algo) {
   algos.push_back(algo);
}

void Moveable::removeAlgo(const MoveableAlgoPtr &algo) {
   AlgoVec::iterator iter;
   for (iter = algos.begin(); iter != algos.end(); ++iter)
      if (*iter == algo)
         break;
   if (iter == algos.end())
      throw spug::DoesNotExistException("algo to remove does not exist");
}

void Moveable::addCollidable(const SimpleCollidablePtr &collidable) {
   collidables.push_back(collidable);
   collidable->setMoveable(this);
}

void Moveable::setCollisionAlgo(const CollisionAlgoPtr &algo) {
   collisionAlgo = algo;
}

bool Moveable::checkForCollision(const CollidablePtr &other, Float &distance) {
   if (collisionAlgo) {

      // delegate checking to the collision algorithm
      Vector relativeVelocity = velocity - other->getVelocity();
      relativeVelocity.normalize();
      bool collided = collisionAlgo->checkForCollision(position, 
                                                       relativeVelocity, 
                                                       *other, 
                                                       distance
                                                       );

      // if we collided, bind our maximum distance to the collision
      if (collided && distance < maxDistance) {
	 if (trace) 
	    cout << "got collision, with " << *other 
	       << ", max distance bounded to " << distance << endl;
         setMaxDistance(distance);
         setCollisionTarget(other);
      }

      return collided;
   } else {
      return false;
   }
}

void Moveable::informOfCollision(const CollidablePtr &other, Float &distance) {
   if (distance < maxDistance) {
      if (trace)
         std::cout << "informOfCollision() collided moveable with " << *other
            << std::endl;
      setMaxDistance(distance);
      setCollisionTarget(other);
   }
}

void Moveable::processRenderFrame(GameEngine &engine, Float elapsed) {

   // skip this if we have no further to travel
   if (!maxDistance)
      return;

   // calculate the change in position for the amount of time since the last
   // render.
   Vector displacement = velocity * elapsed;

   // if we are about to exceed the maximum distance, stop there
   Float distance = displacement.size();
   if (distance >= maxDistance) {
      if (trace) 
	 cout << "colliding.  max dist = " << maxDistance << endl;
      displacement.normalize();
      displacement *= maxDistance;
      maxDistance = 0;

      // call the collision handler if there is one
      if (collisionHandler)
         collisionHandler->handleCollision(*collisionTarget);
   } else {
      if (trace) 
	 cout << "traveling " << distance << " units" << endl;
      maxDistance -= distance;
   }

   // move the object
   position += displacement;
}

void Moveable::processEngineFrame(GameEngine &engine) {
   // run the algos
   for (AlgoVec::iterator iter = algos.begin(); iter != algos.end(); ++iter)
      (*iter)->apply(*this, engine);
}
