/*===========================================================================*\

   $Id$

   Copyright (C) 2007 Michael A. Muller

   Permission is granted to use, modify and redistribute this code,
   providing that the following conditions are met:
 
   1) This copyright/licensing notice must remain intact.
   2) If the code is modified and redistributed, the modifications must 
   include documentation indicating that the code has been modified.
   3) The author(s) of this code must be indemnified and held harmless
   against any damage resulting from the use of this code.

   This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
   warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

\*===========================================================================*/

#include "PointCollisionAlgo.h"
#include "Tri.cc"
#include <spug/Logger.h>

spug::Tracer::Level PointCollisionAlgo::collision = 
   spug::Tracer::get().getLevel("PointCollisionAlgo.calculations");

bool PointCollisionAlgo::checkPolygonCollision(const Vector &position,
                                               const Vector &direction,
                                               const Poly &poly,
                                               Float &distance
                                               ) const {
   SPUG_TRACE(collision, "checking for collision with " << poly);
   // break the polygon up into triangles
   size_t sides = poly.getSideCount();
   for (int i = 0; i < sides - 2; ++i) {

      // check for intersections with each of them
      Tri tri(poly.getVertex(0), poly.getVertex(i + 1), poly.getVertex(i + 2));
      Vector intersection;
      if (tri.intersects(position, direction, intersection, distance)) {
         SPUG_TRACE(collision, "  got collision");
         return true;
      }
   }

   SPUG_TRACE(collision, "  no collision");
   return false;
}

bool PointCollisionAlgo::intersectsAxisPlane(Float pos, Float dir, Float min, 
                                             Float max, 
                                             Float &dist
                                             ) const {
   // if the direction is zero, the point will never intersect with either
   // plane
   if (isZero(dir))
      return false;

   if (pos < min) {
      // less than minimum
      if (dir > 0) {
         dist = (min - pos) / dir;
         return true;
      } else
         return false;
   } else if (pos > max) {
      // greater than maximum
      if (dir < 0) {
         // since dir is negative, and we want positive distance, subtract pos
         // from max instead of the opposite
         dist = (max - pos) / dir;
         return true;
      } else
         return false;
   } else {
      // coord currently lies between the minimum and the maximum - we can't
      // use this
      return false;
   }

}

bool PointCollisionAlgo::inRange(Float pos, Float dir, Float dist, Float min,
                                 Float max
                                 ) {
   Float point = pos + dir * dist;
   return point >= min && point <= max;
}

bool PointCollisionAlgo::checkAxisBoxCollision(const Vector &position,
                                               const Vector &direction,
                                               const AxisBox &box,
                                               Float &distance
                                               ) const {

   SPUG_TRACE(collision, "checking for axis box collision with " << box);

   // first check for the trivial condition where the origin is in the box
   if (box.encloses(position)) {
      distance = 0;
      return true;
   }

   // for each axis, see if the ray intersects the planes perpendicular to the
   // axis at a point within the rectangle of the side of the ox at lies on
   // that plane
   Float dist;
   if (intersectsAxisPlane(position.x, direction.x, box.min.x, box.max.x,
                           dist
                           ) &&
       inRange(position.y, direction.y, dist, box.min.y, box.max.y) &&
       inRange(position.z, direction.z, dist, box.min.z, box.max.z)
       ) {
      distance = dist;
      return true;

   } else if (intersectsAxisPlane(position.y, direction.y, box.min.y, 
                                  box.max.y,
                                  dist
                                  ) &&
              inRange(position.x, direction.x, dist, box.min.x, box.max.x) &&
              inRange(position.z, direction.z, dist, box.min.z, box.max.z)
              ) {
      distance = dist;
      return true;
   } else if (intersectsAxisPlane(position.z, direction.z, box.min.z, 
                                  box.max.z,
                                  dist
                                  ) &&
              inRange(position.x, direction.x, dist, box.min.x, box.max.x) &&
              inRange(position.y, direction.y, dist, box.min.y, box.max.y)
              ) {
      distance = dist;
      return true;
   }

   return false;
}

bool PointCollisionAlgo::checkForCollision(const Vector &position, 
                                           const Vector &direction,
                                           const Collidable &other,
                                           Float &distance
                                           ) const {
                               
   // see what target types are supported
   Collidable::Types types = other.getTypes();

   // first check for a polygon
   if (types.includes(Collidable::poly))
      return checkPolygonCollision(position, direction, other.getPoly(), 
                                   distance
                                   );
   // XXX sphere should be easy
   else if (types.includes(Collidable::axisBox))
      return checkAxisBoxCollision(position, direction, other.getAxisBox(), 
                                   distance
                                   );
   else {
      SPUG_WARN("Can not check for collision with object.");
      return false;
   }
}

